(*SEMANTIC EVALUATION FUNCTIONS  *)

let rec makefun ((a:exp),(x:dval env)) =
      (match a with
      |	Fun(ii,aa) ->
      	  Dfunval(function (d,s,h) ->  sem aa (bindlist(x, ii, d)) s h)
      |	_ -> failwith ("Non-functional object"))
      
and makeproc ((a:exp),(x:dval env)) =
    (match a with
      |	Proc(ii,b) ->
	  Dprocval(function (d1,s1, h1) ->
			 semb b (bindlist(x, ii, d1)) s1 h1) 
      |	_ -> failwith ("Non-procedural object"))
      
and applyfun ((ev1:dval),(ev2:dval list),s, h) =
      ( match ev1 with
      | Dfunval(x) -> x (ev2,s, h)
      | _ -> failwith ("attempt to apply a non-functional object"))
      
and applyproc ((ev1:dval),(ev2:dval list),s, h) =
      ( match ev1 with
      | Dprocval(x) -> x (ev2,s, h)
      | _ -> failwith ("attempt to apply a non-functional object"))
      
and makefunrec (i, Fun(ii, aa), r) = 
    let functional ff (d,s1, h1) = 
	     let r1 = bind(bindlist(r, ii, d), i, Dfunval(ff)) in
	     sem aa r1 s1 h1 in 
       let rec fix =  function x -> functional fix x 
          in Funval(fix) 

and makeclass((c: cdecl), r) = match c with
  Class(name, fpars, extends, (b1,b2,b3) ) ->
    Classval(function  (apars, s, h) ->
	 (match extends with
	   | ("Object",_) -> 
		let i = newpoint() in 
		(let (r2, s2, h2) =
		 semdl (b1, b2) 
		   (bindlist(r, fpars @ ["this"], apars @ [Dobject(i)])) s h in 
		 let (s3, h3) = semcl b3 r2 s2 h2 in
		 let r3 = localenv( r2, fpars, r) in  
		 let newh = allocateheap(h3, i, r3) in (Object i, s3, newh ))
	   | (super,superpars) ->  
		 let (v, s1, h1) = 
		 applyclass(applyenv(r,super), 
		   findsuperargs( fpars, apars, superpars), s, h) in
		   let n = (match v with | Object n1 -> n1) in
		   let (r2, s2, h2) =  semdl (b1, b2)  
		    (bindlist(eredita( r, v, h1), fpars, apars)) s1 h1 in
		   let (s3, h3) = semcl b3 r2 s2 h2 in
		   let newh = allocateheap(h3, n, localenv( r2 ,fpars, r)) in
		    (Object n, s3, newh) 
			      ))	  
and applyclass ((ev1:dval),(ev2:dval list), s, h) =
  ( match ev1 with
	| Classval(x) -> x (ev2, s, h)
	| _ -> failwith ("notaclass"))
	
and sem (e:exp) (r:dval env) (s: mval store) (h:heap) =
      match e with
      | Eint(n) -> (Int(n),s,h)
      | Ebool(b) -> (Bool(b),s,h)
      | Den(i) -> (dvaltoeval(applyenv(r,i)),s,h)
      | Iszero(a) -> let (v1,s1,h1) = sem a r s h in (iszero(v1),s1,h1)
      | Eq(a,b) -> let (v1,s1,h1) = sem a r s h in
	               let (v2,s2,h2) = sem b r s1 h1 in
				   (equ(v1 ,v2),s2,h2)
      | Prod(a,b) ->  let (v1,s1,h1) = sem a r s h in
	               let (v2,s2,h2) = sem b r s1 h1 in
				   (mult(v1 ,v2),s2,h2)
      | Sum(a,b) ->  let (v1,s1,h1) = sem a r s h in
	               let (v2,s2,h2) = sem b r s1 h1 in
				   (plus(v1 ,v2),s2,h2)
      | Diff(a,b)  ->  let (v1,s1,h1) = sem a r s h in
	               let (v2,s2,h2) = sem b r s1 h1 in
				   (diff(v1 ,v2),s2,h2)
      | Minus(a) ->  let (v1,s1,h1) = sem a r s h in (minus(v1),s1,h1)
      | And(a,b) ->  let (v1,s1,h1) = sem a r s h in
	               let (v2,s2,h2) = sem b r s1 h1 in
				   (et(v1 ,v2),s2,h2)
      | Or(a,b) ->  let (v1,s1,h1) = sem a r s h in
	               let (v2,s2,h2) = sem b r s1 h1 in
				   (vel(v1 ,v2),s2,h2)
      | Not(a) -> let (v1,s1,h1) = sem a r s h in (non(v1),s1,h1)
      | Ifthenelse(a,b,c) -> 
           let (g,s1,h1) = sem a r s h in
		  if typecheck("bool",g) then
		     (if g = Bool(true) 
		     then sem b r s1 h1
		     else sem c r s1 h1)
		  else failwith ("nonboolean guard")            
      | Fun(i,a) ->  (dvaltoeval(makefun(Fun(i,a), r)), s, h)
      | Appl(a,b) -> let (v1,s1,h1) = semden a r s h in
	    let (v2,s2,h2) = semlist b r s1 h1 in
				   applyfun(v1, v2, s2, h2)
      | Let(i,e1,e2) -> let (v, s1, h1) = semden e1 r s h in 
	      sem e2 (bind(r, i, v)) s1 h1
      | Rec(f,e) -> (makefunrec (f,e,r), s, h)
      | Val(e) -> let (v, s1, h1) = semden e r s h in
	      (match v with 
		   | Dloc n -> (mvaltoeval(applystore(s1, n)), s1, h1)
		   | _ -> failwith("not a variable"))
      | New(i,ge) -> let (v, s1, h1) = semlist ge r s h in 
           applyclass(applyenv(r,i), v, s1, h1)
      | This -> (dvaltoeval(applyenv(r,"this")), s, h) 
	  | _ -> failwith ("nonlegal expression for sem") 
      
and semlist el r s h = match el with
	  | [] -> ([], s, h)
	  | e::el1 -> let (v1, s1, h1) = semden e r s h in 
	    let (v2, s2, h2) = semlist el1 r s1 h1 in
	     (v1 :: v2, s2, h2)
		 
and semden (e:exp) (r:dval env) (s: mval store) (h:heap) =  match e with
      | Den(i) -> (applyenv(r,i), s, h)
      | Fun(i,e1) -> (makefun(e,r), s, h)
      | Proc(il,b) -> (makeproc(e,r), s, h)
      | Newloc(e) -> let (v, s1, h1) = sem e r s h in
             let m = evaltomval v in let (l, s2) = allocate(s1, m)
	      in (Dloc l, s2, h1)
      | Field(e,i) -> (match sem e r s h with
		  | (Object i1,s1,h1) -> let r1 = applyheap(h1, i1) in 
		        let field = applyenv(r1,i) in
				 (field,s1,h1)				 
		  | _ -> failwith("notanobject"))
      | _ -> let (v, s1, h1) = sem e r s h in
            (evaltodval(v), s1, h1)
	  
and semdv dl r s h = 
    match dl with
	   | [] -> (r,s,h)
	   | (i,e)::dl1 -> let (v, s1, h1) = semden e r s h in
	        semdv dl1 (bind(r, i, v)) s1 h1
and semdl (dl,rl) r s h =
    let (r1, s1, h1) = semdv dl r s h in
     semdr rl r1 s1 h1
    
and semdr rl r s h =    
    let functional ((r1: dval env)) = 
       (match rl with
	   | [] -> r
	   | (i,e) :: rl1 -> let (v, s2, h2) = semden e r1 s h in
	         let (r2, s3, h3) = semdr rl1 (bind(r, i, v)) s h in
		 r2) in
     let rec rfix = function x -> functional rfix  x in
     (rfix, s, h)
	 
and semc (c: com) (r:dval env) (s: mval store) (h:heap) = match c with
	
	   | Assign(e1, e2) -> let (v1, s1, h1) = semden e1 r s h in
	      (match v1 with
	     | Dloc(n) -> let (v, s2, h2) = sem e2 r s1 h1 in
	      (update(s2, n, evaltomval v), h2)
	     | _ -> failwith ("wrong location in assignment"))
		 
	   | Cifthenelse(e, cl1, cl2) -> let (g, s1, h1) = sem e r s h in
              if typecheck("bool",g) then
               (if g = Bool(true) 
               then semcl cl1 r s1 h1
               else semcl cl2 r s1 h1)
              else failwith ("nonboolean guard")
	      
	   | While(e, cl) -> 
     	    let functional ((fi: (mval store) * heap -> (mval store) * heap)) = 
	      function (sigma, delta) ->
	       let (g, s1, h1) = sem e r sigma delta in
                if typecheck("bool",g) then
                   (if g = Bool(true) 
		   then fi(semcl cl  r sigma delta)  
		   else (sigma,delta))
	        else failwith ("nonboolean guard") in 
	      let rec statefix = function x -> functional statefix x in
	      (match statefix(s, h) with | (sfix, hfix) ->
			   (sfix, hfix))
			
	   | Call(e1, e2) -> let (p, s1, h1) = semden e1 r s h  in
	      let (v, s2, h2) = semlist e2 r s1 h1 in 
	      applyproc(p, v, s2, h2) 
		  
	   | Block(b) -> semb b r s h
			
	     
and semcl cl r s h = match cl with
	   | [] -> s, h
	   | c::cl1 -> let (s1,h1) = (semc c r s h ) in semcl cl1 r s1 h1
	   
and semb (dl, rdl, cl) r s h = 
	  let (r1, s1, h1) = semdl (dl,rdl) r s h in
	       semcl cl r1 s1 h1
	   
and semclasslist cl ( r: dval env )  = 
     let functional (r1: dval env) =
       (match cl with
	  | [] -> r
	  | Class(nome,x1,x2,x3)::cl1 -> semclasslist cl1 (bind(r,nome, 
	      makeclass(Class(nome,x1,x2,x3), r1)))) in 
     let rec rfix = function i -> functional rfix i 
     in rfix
	  
and semprog (cdl,b) r s h = semb b (semclasslist cdl r) s h



