(* SEMANTIC DOMAINS *)

exception Nonstorable
exception Nonexpressible
type pointer = int 
type 't env = Denv of int | Penv of pointer  (* cambiato *)
and loc = mval store * int (* cambiato *)
and 't store =  Dstore of int | Pstore of pointer (* cambiato *) 
and eval =  	   
	| Int of int 
	| Bool of bool
	| Novalue
	| Funval of efun 
	| Object of pointer 
	
and efun =  exp * (dval env) 
and proc =  exp * (dval env) 
and eclass = cdecl * (dval env) 

		
and obj = ( ide array * dval array * dval env * mval array )  (* cambiato *)  	
and heap = obj array  (* cambiato *)

and dval =  	   
	| Dint of int 
	| Dbool of bool 
	| Unbound
	| Dloc of loc
	| Dfunval of efun
	| Dprocval of proc 
	| Dobject of pointer 
	| Classval of eclass 
	
and mval = 
	| Mint of int
	| Mbool of bool
	| Undefined
	| Mobject of pointer 
	
let evaltomval e = match e with
	| Int n -> Mint n
	| Bool n -> Mbool n
	| Object n -> Mobject n 
	| _ -> raise Nonstorable
	
let mvaltoeval m = match m with
	| Mint n -> Int n
	| Mbool n -> Bool n
	| Mobject n -> Object n 
	| _ -> Novalue
	
let evaltodval e = match e with
	| Int n -> Dint n
	| Bool n -> Dbool n
	| Novalue -> Unbound
	| Funval f -> Dfunval f
	| Object n -> Dobject n
	
let dvaltoeval e = match e with
	| Dint n -> Int n
	| Dbool n -> Bool n
	| Dloc n -> raise Nonexpressible
	| Dprocval n -> raise Nonexpressible
	| Unbound -> Novalue
	| Dfunval f -> Funval f
	| Dobject n -> Object n
	| Classval n ->  raise Nonexpressible

type labeledconstruct = 
   | Expr1 of exp 
   | Expr2 of exp
   | Exprd1 of exp
   | Exprd2 of exp
   | Com1 of com
   | Com2 of com
   | Coml of labeledconstruct list
   | Decl of labeledconstruct list
   | Rdecl of (ide * exp) list
   | Dec1 of (ide * exp) 
   | Dec2 of (ide * exp) 
   | Ogg1 of cdecl  
   | Ogg2 of cdecl 
   | Ogg3 of cdecl 
   
let (cstack: labeledconstruct stack stack) = 
  emptystack(stacksize,emptystack(1,Expr1(Eint(0))))

let (tempvalstack: eval stack stack) = emptystack(stacksize,emptystack(1,Novalue))

let (tempdvalstack: dval stack stack) = emptystack(stacksize,emptystack(1,Unbound))

let labelcom (dl: com list) = let dlr = ref(dl) in
	let ldlr = ref([]) in
	while not (!dlr = []) do
	 let i = List.hd !dlr in
	 ldlr := !ldlr @ [Com1(i)];
	 dlr := List.tl !dlr
	done;
	Coml(!ldlr)
	
let labeldec (dl: (ide * exp) list) = let dlr = ref(dl) in
	let ldlr = ref([]) in
	while not (!dlr = []) do
	 let i = List.hd !dlr in
	 ldlr := !ldlr @ [Dec1(i)];
	 dlr := List.tl !dlr
	done;
	Decl(!ldlr)
	 
let (labelstack: labeledconstruct stack) =
	  emptystack(stacksize,Expr1(Eint(0))) 


let pushargs ((b: exp list),(continuation: labeledconstruct stack)) = 
     let br = ref(b) in
      while not(!br = []) do
	    push(Exprd1(List.hd !br),continuation);
	    br := List.tl !br
	  done

let getargs ((b: exp list),(tempstack: dval stack)) = 
      let br = ref(b) in
      let er = ref([]) in
      while not(!br = []) do
	    let arg=top(tempstack) in
	pop(tempstack); er := !er @ [arg];
	    br := List.tl !br
	  done; 
	  !er
	  

(* implementazione heap *)
let (newpoint,initpoint) =
   let count = ref(-1) in
   (fun () -> count := !count +1;
	      !count),
   (fun () -> count := -1)
   

let emptyheap () = initpoint(); 
      (Array.create 100 ((Array.create 1 "dummy"),(Array.create 1 Unbound),
      Denv(-1), (Array.create 1 Undefined))
         : heap) 

let currentheap = ref(emptyheap())	 
	 
let applyheap ((x: heap), (y:pointer)) = Array.get x y

let allocateheap ((x:heap), (i:pointer), (r:obj)) = 
    Array.set x i r; x
    
let getnomi(x:obj) = match x with
  | (nomi, den, slink, st) -> nomi
  
let getden(x:obj) = match x with
  | (nomi, den, slink, st) -> den
  
let getslink(x:obj) = match x with
  | (nomi, den, slink, st) -> slink
  
let getst(x:obj) = match x with
  | (nomi, den, slink, st) -> st
(* ambiente: cambiato *)
let currentenv = ref(0)
 
let namestack = emptystack(stacksize,[| "dummy" |])

let dvalstack = emptystack(stacksize,[| Unbound |])

let slinkstack = emptystack(stacksize, Denv(0))

let overriden ((a: ide array), (i: ide)) =
    let len = Array.length a in
    let ind = ref(0) in
    let res = ref(false) in
    while !ind < len do
      if Array.get a !ind = i 
       then (res := true; ind := len)
       else ind := !ind + 1
    done;
    !res 
let bindlist(r, il, el) = 
   let n = List.length il in
   let ii = Array.create n "dummy" in
   let dd = Array.create n Unbound in
   let ri = ref(il) in
   let rd = ref(el) in
   let index = ref(0) in
   while !index < n do
     let i = List.hd !ri in
	 let d = List.hd !rd in
	 ( ri := List.tl !ri;
	   rd := List.tl !rd;
	   Array.set ii !index i;
	   Array.set dd !index d;
	   index := !index + 1)
    done;
    push(ii, namestack);
    push(dd,dvalstack);
    push(r,slinkstack);
    (Denv(lungh(namestack)): dval env)  (* qui *)
	
let bind ((x:dval env),i,d) =  (* solo il prologo *) 
   match x  with
      | Penv(x1) -> failwith("bind for objects")
      | Denv(r) -> 
	let arrnomi = access(namestack, r) in
	let lun =  Array.length arrnomi in
	let arrden = access(dvalstack, r) in
	let n = ref(0) in
	let index = ref(-1) in
	while !n < lun do
	 let nome = Array.get arrnomi !n in 
	 if nome = i then 
	   (Array.set arrden !n d;
	   n := lun; index := -2) else  
	 if nome = "dummy" 
	 then (index := !n; n := lun)
	 else n := !n + 1
	done;
	if !index = -2 then x else 
	if !index > -1 
	then
	   (Array.set arrnomi !index i;
	   Array.set arrden !index d; x)
	else 
	   (push(Array.create 1 i,namestack); 
	   push(Array.create 1 d,dvalstack);
	   push(x,slinkstack);
	   (Denv(lungh(namestack)): dval env))  (* qui *)
	   

let emptyenv(x) = currentenv := -1; svuota(namestack);
   svuota(dvalstack); svuota(slinkstack); 
   Denv(!currentenv)  (* qui *)

let svuotaenv() = svuota(namestack); 
    svuota(dvalstack); svuota(slinkstack)



let applyenv ((r: dval env), (y: ide)) =  (* nuova *)
    let den = ref(Unbound) in
    let (x, caso) = (match r  with
      | Denv(x1) -> (x1, ref("stack"))
      | Penv(x1) -> (x1, ref("heap"))) in
       let n = ref(x) in
       while !n > -1 do
	  let lenv = 
	  if !caso = "stack" 
	    then access(namestack,!n) 
	    else getnomi(applyheap(!currentheap,!n)) in
	      let nl = Array.length lenv in
	      let index = ref(0) in
	      while !index < nl do
                if Array.get lenv !index = y then
		 (den := (if !caso = "stack" 
		 then Array.get (access(dvalstack,!n)) !index
		 else Array.get (getden(applyheap(!currentheap,!n))) !index);
		 index := nl)
		else index := !index + 1
	      done;
           if not(!den = Unbound) then  n := -1
	   else  let next =  (if !caso = "stack" then access(slinkstack,!n)
	      else getslink(applyheap(!currentheap,!n))) in
	      caso := (match next with
	        |Denv(_) -> "stack"
	        |Penv(_) -> "heap");
	      n := (match next with
	        |Denv(n1) -> n1
	        |Penv(n1) -> n1)	      
       done;
       !den;;
	  


	   

	


let pushlocalenv (dl,dlr,x) = 
   match x with
      | Penv(r) -> failwith("pushlocalenv for objects")
      | Denv(r) ->
   let anomi = access(namestack, r) in
   let aden = access(dvalstack, r) in
   let nn = ref(Array.length anomi) in
   let lun1 = ref(0) in
   let dlp = ref(dl) in
   while not(!dlp = []) do
      let i = (match List.hd !dlp with | (i1, _) -> i1) in
      dlp := List.tl !dlp;
      if overriden(anomi, i) then failwith("field overriding is not allowed")
      else lun1:= !lun1 + 1
   done;
   let lun2 = ref(0) in
   let dlp = ref(dlr) in
   while not(!dlp = []) do
      let i = (match List.hd !dlp with | (i1, _) -> i1) in
      dlp := List.tl !dlp;
      if overriden(anomi, i) then ()
      else lun1:= !lun1 + 1
   done;
   let rn = !lun1 + !lun2 + !nn in
   let ii = Array.create rn "dummy" in
   let dd = Array.create rn Unbound in
   while not(!nn = 0) do
      Array.set ii (!nn - 1) (Array.get anomi (!nn - 1));
      Array.set dd (!nn - 1) (Array.get aden (!nn - 1));
      nn := !nn - 1
   done;
   pop(namestack); pop(dvalstack);
   push(ii, namestack);
   push(dd, dvalstack);
   x


let topenv() = Denv(!currentenv)

let popenv () = pop(namestack);
	  pop(dvalstack);
	  pop(slinkstack);
	  currentenv := !currentenv - 1
	  
let pushemptylocalenv(r) = push([||],namestack);
	   push([||],dvalstack);
	   push(r,slinkstack);
	   currentenv := lungh(namestack)
	   
let pushenv(r1) = match r1 with Denv(r) ->
     if r = !currentenv then 
	   pushemptylocalenv(r1)
	 else currentenv := r
	    





let currentstore = ref(0)

let storestack = emptystack(stacksize,[|Undefined|])

(* tutte nuove *)
     
let (newloc,initloc) =
   let count = ref(-1) in
   (fun () -> let arrst = access(storestack, !currentstore) in
              let len = Array.length arrst in
	      let index = ref(0) in
	      while !index < len do
	        if Array.get arrst !index = Undefined then
		(count := !index; index := len)
		else index := !index + 1
	      done;
	      (Dstore(!currentstore), !count)),
   (fun () -> count := -1)

let applystore ((x: mval store), (d: loc)) = 
   match d with 
     | (s2,n2)->
       (match s2 with
      | Dstore(n1) -> let a = access(storestack, n1) in
	           Array.get a n2
      | Pstore(n1) -> let a = getst(applyheap(!currentheap,n1)) in  
                   Array.get a n2)
     | _ -> failwith("not a location in applystore")
       
	   
let emptystore(x) = initloc(); 
   svuota(storestack); currentstore := -1; Dstore(!currentstore) 

let allocate ((s:mval store),  (m:mval)) = 
    let (s2, n2) = newloc() in 
    (match s2 with
      | Pstore(n1) -> let a = access(storestack, n1) in
	           Array.set a n2 m; ((s2, n2),s)
      | Dstore(n1) -> let a = access(storestack, n1) in 
                   Array.set a n2 m;  ((s2, n2),s))
     
let update((s:mval store),  (d: loc),  (m:mval)) =
    if applystore(s, d) = Undefined 
         then failwith ("wrong assignment")
     else  match d with 
     | (s2,n2) ->
      (match s2 with
      | Dstore(n1) -> let a = access(storestack, n1) in
	           Array.set a n2 m; s
      | Pstore(n1) -> let a = getst(applyheap(!currentheap,n1)) in
                   Array.set a n2 m;  s )
     
let pushlocalstore (dl) = 
   let currarr = access(storestack, !currentstore) in
   let currl = Array.length currarr in
   let rdl = ref(dl) in
   let rn = ref(0) in
   while not(!rdl = []) do
      let (i, d) = List.hd !rdl in
	  (match d with
	    | Newloc(_) -> rn := !rn + 1
	    | _ -> ());
      rdl := List.tl !rdl
   done;
   let a = Array.create (!rn + currl) Undefined in
   let index1 = ref(0) in
   while !index1 < currl do
     let y = Array.get currarr !index1 in 
     Array.set a !index1 y; 
     index1 := !index1 + 1
   done;
   pop(storestack); push(a, storestack); 
   Dstore(!currentstore)

let pushemptylocalstore () = push([||],storestack); 
   currentstore := !currentstore + 1

let pushstore(n1) = match n1 with Dstore(n) ->
    if n = !currentstore 
        then pushemptylocalstore()
	else currentstore := n

let popstore () = pop(storestack); currentstore := !currentstore -1

let svuotastore () = svuota(storestack)

let topstore() = Dstore(!currentstore)


let dlist ((b: ide list),(r: dval env)) = 
      let br = ref(b) in
      let er = ref([]) in
      while not(!br = []) do
	    let arg= applyenv(r, List.hd !br) in
	    er := !er @ [arg];
	    br := List.tl !br
	  done;
	  !er

    
let notoccur ((i:ide), l) = let vl = ref(l) in
      let res = ref(true) in
      while not (!vl = []) do
	 if i = List.hd(!vl) then (res := false; vl := [])
	     else vl := List.tl(!vl)
      done;
      !res
      
let findone ((args: ide list), (el: dval list), (s: ide )) =   
          let vargs = ref(args) in
	  let dargs = ref(el) in
	      let valore = ref(Unbound) in
	      while not(!vargs = [])  do 
		if s = List.hd !vargs then 
		(vargs := []; valore := List.hd !dargs) 
		else  vargs := List.tl !vargs; dargs := List.tl !dargs
	      done;
	      !valore      
let findsuperargs ((args: ide list) ,(el: dval list) , (sargs: ide list)) = 
   let vsargs = ref(sargs) in
   let res = ref([]) in
   while not(!vsargs = []) do
     let i = findone(args, el, List.hd(!vsargs)) in
     if i = Unbound then failwith("problem with super parameters")
     else  res := !res @ [i]; vsargs := List.tl !vsargs
   done;
   !res
   
   (* da rivedere *)
let localenv ((r:dval env) , (s: mval store), Dobject(ob), (li:ide list), 
(r1: dval env)) =
    let (rint, sint) = (match (r,s) with
      | (Denv nr, Dstore ns) -> (nr, ns)
      | _ -> failwith("heap structures in localenv")) in
    let oldst = access(storestack, sint) in
    let oldnomi = access(namestack, rint) in
    let oldden = access(dvalstack, rint) in
    let storesize = Array.length oldst in
    let newst = Array.create storesize Undefined in
    let index = ref(0) in
    while not(!index = storesize) do
       Array.set newst !index (Array.get oldst !index);
       index := !index + 1
    done;
    let oldenvlength = Array.length oldnomi in
    let newenvlength = oldenvlength - (List.length li) in
    let newnomi = Array.create newenvlength "dummy" in
    let newden = Array.create newenvlength Unbound in
    let index = ref(0) in
    let newindex = ref(0) in
    while not(!index = oldenvlength) do
      let lname = Array.get oldnomi !index in
      let lden = Array.get oldden !index in
      if notoccur(lname, li) then (
      Array.set newnomi !newindex lname;
      let ldennuova = (match lden with
        | Dfunval(e,rho) -> if rho >= r then Dfunval(e,Penv(ob)) else lden
	| Dprocval(e,rho) -> if rho >= r then Dprocval(e,Penv(ob)) else lden
	| Dloc(sigma, n) -> if sigma >= s then Dloc(Pstore(ob),n) else lden
	|  _ -> lden              ) in
      Array.set newden !newindex ldennuova;
      newindex := !newindex + 1)
      else ();
      index := !index + 1
    done;
    (newnomi, newden, r1, newst)
	
let eredita ((rho: dval env), ogg, (h: heap)) =
   let currn = (match rho with | Denv(n) -> n) in
   let (point, arrnomisotto, arrdensotto, arrstore) =
   (match ogg with
     | Object(n) -> let oo = applyheap(!currentheap,n) in
       (n, getnomi(oo), getden(oo), getst(oo))
     | _ -> failwith("not an object in eredita")) in
   (* store *)
   let stl = Array.length arrstore in
   let newstore = Array.create stl Undefined in
   let index = ref(0) in
   while !index < stl do
     Array.set newstore !index (Array.get arrstore !index);
     index := !index + 1
   done;
   pop(storestack);
   push(newstore, storestack);
   (* env *)
   let currarrnomi = access(namestack, currn) in
   let currarrden = access(dvalstack, currn) in
   let r = access(slinkstack, currn) in
   let currl = Array.length currarrnomi in
   let oldlen = Array.length arrnomisotto in
   index := 0;
   while not(Array.get arrnomisotto !index = "this") do  
    index := !index + 1 done;
   index := !index + 1;
   let newlen = (currl + oldlen - !index )  in
   let newarrnomi = Array.create newlen  "dummy" in 
   let newarrden = Array.create newlen  Unbound in
   let newindex = ref(0) in
   while !newindex < currl do
     Array.set newarrnomi !newindex (Array.get currarrnomi !newindex);
     Array.set newarrden !newindex (Array.get currarrden !newindex);
     newindex := !newindex + 1
   done;
   while !newindex < newlen do
     Array.set newarrnomi !newindex (Array.get arrnomisotto !index);
     Array.set newarrden !newindex (Array.get arrdensotto !index);
     newindex := !newindex + 1;
     index := !index + 1
   done;
   pop(namestack);pop(dvalstack);pop(slinkstack);
   push(newarrnomi, namestack);
   push(newarrden,dvalstack);
   push(r,slinkstack);



