module Sem

(*SEMANTIC EVALUATION FUNCTIONS  *)

let rec sem ((e:exp), (r:dval env), (s: mval store)) =
  match e with
    | Eint(n) -> Int(n)
    | Ebool(b) -> Bool(b)
    | Den(i) -> dvaltoeval(applyenv(r,i))
    | Iszero(a) -> iszero(sem(a, r, s))
    | Eq(a,b) -> equ(sem(a, r, s), sem(b, r, s))
    | Prod(a,b) ->  mult (sem(a, r, s), sem(b, r, s))
    | Sum(a,b) ->  plus (sem(a, r, s), sem(b, r, s))
    | Diff(a,b)  ->  diff (sem(a, r, s), sem(b, r, s))
    | Minus(a) ->  minus(sem(a, r, s))
    | And(a,b) ->  et (sem(a, r, s), sem(b, r, s))
    | Or(a,b) ->  vel (sem(a, r, s), sem(b, r, s))
    | Not(a) -> non(sem(a, r, s))
    | Ifthenelse(a,b,c) ->
        let g = sem(a, r, s)
        if typecheck("bool",g) then
          if g = Bool(true) then
            sem(b, r, s)
          else
            sem(c, r, s)
        else
          failwith ("nonboolean guard")
    | Let(i,e1,e2) ->
        let (v, s1) = semden(e1, r, s)
        sem(e2, bind (r ,i, v), s1)
    | Val(e) ->
        let (v, s1) = semden(e, r, s)
        match v with
          | Dloc n -> mvaltoeval(applystore(s1, n))
          | _ -> failwith("not a variable")
    | _ -> failwith ("nonlegal expression for sem")

and semden ((e:exp), (r:dval env), (s: mval store)) =
  match e with
  | Den(i) -> (applyenv(r,i), s)
  | Newloc(e) ->
      let m = evaltomval(sem(e, r, s))
      let (l, s1) = allocate(s, m)
      (Dloc l, s1)
  | _ -> (evaltodval(sem(e, r, s)), s)

and semdv(dl, r, s) =
  match dl with
    | [] -> (r,s)
    | (i,e)::dl1 ->
        let (v, s1) = semden(e, r, s)
        semdv(dl1, bind(r, i, v), s1)

and semc((c: com), (r:dval env), (s: mval store)) =
  match c with
    | Assign(e1, e2) ->
        let (v1, s1) = semden(e1, r, s)
        match v1 with
          | Dloc(n) -> update(s1, n, evaltomval(sem(e2, r, s)))
          | _ -> failwith ("wrong location in assignment")
    | Cifthenelse(e, cl1, cl2) ->
        let g = sem(e, r, s)
        if typecheck("bool",g) then
          if g = Bool(true) then
            semcl(cl1, r, s)
          else
            semcl (cl2, r, s)
        else
          failwith ("nonboolean guard")
    | While(e, cl) ->
        let g = sem(e, r, s)
        if typecheck("bool",g) then
          if g = Bool(true) then
            semcl((cl @ [While(e, cl)]), r, s)
          else
            s
        else failwith ("nonboolean guard")
    | Block(b) -> semb(b, r, s)

and semcl(cl, r, s) =
  match cl with
    | [] -> s
    | c::cl1 -> semcl(cl1, r, semc(c, r, s))

and semb ((dl, cl), r, s) =
  let (r1, s1) = semdv(dl, r, s)
  semcl(cl, r1, s1)
