(*SEMANTIC EVALUATION FUNCTIONS  *)




let itsem(rho) =
    (match top(continuation) with
     |Expr1(x) -> 
         (pop(continuation);
         push(Expr2(x),continuation); 
         (match x with
      
          | Iszero(a) -> push(Expr1(a),continuation)
          | Eq(a,b) -> push(Expr1(a),continuation); 
                push(Expr1(b),continuation)
	  | Prod(a,b) ->  push(Expr1(a),continuation); 
                push(Expr1(b),continuation)
          | Sum(a,b) -> push(Expr1(a),continuation); 
                push(Expr1(b),continuation)
          | Diff(a,b)  ->  push(Expr1(a),continuation); 
                push(Expr1(b),continuation)
          | Minus(a) ->  push(Expr1(a),continuation)
          | And(a,b) ->  push(Expr1(a),continuation); 
                push(Expr1(b),continuation)
          | Or(a,b) ->  push(Expr1(a),continuation); 
                push(Expr1(b),continuation)
          | Not(a) -> push(Expr1(a),continuation)
	  | Ifthenelse(a,b,c) -> push(Expr1(a),continuation)
	  | Val(e) ->  push(Exprd1(e),continuation)
	  | _ -> ()))
    |Expr2(x) -> 
         (pop(continuation); 
         (match x with
          | Eint(n) ->  push(Int(n),tempstack)
          | Ebool(b) -> push(Bool(b),tempstack)
          | Den(i) ->   push(dvaltoeval(applyenv(rho,i)),tempstack)
	  | Iszero(a) -> 
             let arg=top(tempstack) in
             pop(tempstack);
             push(iszero(arg),tempstack)
          | Eq(a,b) -> 
             let firstarg=top(tempstack) in
             pop(tempstack);
             let sndarg=top(tempstack) in
             pop(tempstack);
             push(equ(firstarg,sndarg),tempstack)
	  | Prod(a,b) -> 
             let firstarg=top(tempstack) in
             pop(tempstack);
             let sndarg=top(tempstack) in
             pop(tempstack);
             push(mult(firstarg,sndarg),tempstack)  
          | Sum(a,b) -> 
             let firstarg=top(tempstack) in
             pop(tempstack);
             let sndarg=top(tempstack) in
             pop(tempstack);
             push(plus(firstarg,sndarg),tempstack)
          | Diff(a,b)  -> 
             let firstarg=top(tempstack) in
             pop(tempstack);
             let sndarg=top(tempstack) in
             pop(tempstack);
             push(diff(firstarg,sndarg),tempstack) 
          | Minus(a) -> 
             let arg=top(tempstack) in
             pop(tempstack);
             push(minus(arg),tempstack)  
          | And(a,b) -> 
             let firstarg=top(tempstack) in
             pop(tempstack);
             let sndarg=top(tempstack) in
             pop(tempstack);
             push(et(firstarg,sndarg),tempstack)  
          | Or(a,b) -> 
             let firstarg=top(tempstack) in
             pop(tempstack);
             let sndarg=top(tempstack) in
             pop(tempstack);
             push(vel(firstarg,sndarg),tempstack)  
          | Not(a) -> 
             let arg=top(tempstack) in
             pop(tempstack);
             push(non(arg),tempstack) 
          | Ifthenelse(a,b,c) ->
             let arg=top(tempstack) in
             pop(tempstack);
             if typecheck("bool",arg) then
               (if arg = Bool(true) 
               then push(Expr1(b),continuation)
               else push(Expr1(c),continuation))
            else failwith ("type error") 
	  | Val(e) -> let v = top(tempdstack) in
	      pop(tempdstack);
	      (match v with 
		   | Dloc n -> 
		     push(mvaltoeval(applystore(!globalstore, n)), tempstack)
		   | _ -> failwith("not a variable"))))
          | _ -> failwith("no more cases for semexpr")) 
	  
let itsemden(rho) = 
    (match top(continuation) with
     |Exprd1(x) -> (pop(continuation); push(Exprd2(x),continuation);
       match x with
          | Den i -> ()
	  | _ -> push(Expr2(x), continuation))          
     |Exprd2(x) -> (pop(continuation); match x with
           | Den i -> push(applyenv(rho,i), tempdstack)
	   | _ -> let arg = top(tempstack) in pop(tempstack);
	      push(evaltodval(arg), tempdstack))
     | _ -> failwith("No more cases for semden")    )
	      
    
let itsemcl((rho: dval env)) =
  let cl =
    (match top(continuation) with
      | Coml(dl1) -> dl1
      | _ -> failwith("impossible in semdecl")) in
    if cl = [] then pop(continuation) else
    (let currc = List.hd cl in
    let newcl = List.tl cl in
    pop(continuation); push(Coml(newcl),continuation);
  	(match currc with
  	   | Com1(Assign(e1, e2)) ->  pop(continuation); 
		push(Coml(Com2(Assign(e1, e2))::newcl),continuation);
		push(Exprd1(e1), continuation);
		push(Expr1(e2), continuation)		
	   | Com2(Assign(e1, e2)) -> 
	      let arg2 = evaltomval(top(tempstack)) in
	      pop(tempstack); 
	      let arg1 = top(tempdstack) in
	      pop(tempdstack); 
	      (match arg1 with
  	          | Dloc(n) -> 
		     globalstore := update(!globalstore, n, arg2)				   
  		  | _ -> failwith ("wrong location in assignment"))		  
	   | Com1(While(e, cl)) ->
		pop(continuation); 
		push(Coml(Com2(While(e, cl))::newcl),continuation);
		push(Expr1(e), continuation)	
	   | Com2(While(e, cl)) -> 
	         let g = top(tempstack) in
		 pop(tempstack);
                 if typecheck("bool",g) then
                   (if g = Bool(true) then 
  		     (let old = newcl in 
		     let newl = 
		       (match labelcom cl with
		        | Coml newl1 -> newl1
			| _ -> failwith("impossible in while")) in
  		      let nuovo = Coml(newl @ [Com1(While(e, cl))] @ old) in
  		      pop(continuation); push(nuovo,continuation))
  		   else ())
  		 else failwith ("nonboolean guard")		 
	   | Com1(Cifthenelse(e, cl1, cl2)) -> 
		pop(continuation); 
		push(Coml(Com2(Cifthenelse(e, cl1, cl2))::newcl),continuation);
		push(Expr1(e), continuation)		 
	   | Com2(Cifthenelse(e, cl1, cl2)) -> 
	         let g = top(tempstack) in
		 pop(tempstack);
                 if typecheck("bool",g) then
                    (let temp = if g = Bool(true) then
		    labelcom (cl1) else labelcom (cl2) in
		    let newl = 
		       (match temp with
		        | Coml newl1 -> newl1
			| _ -> failwith("impossible in cifthenelse")) in
		    let nuovo = Coml(newl @ newcl) in
  			   pop(continuation); push(nuovo,continuation))
                else failwith ("nonboolean guard")	
	   | _ -> failwith("no more sensible cases in commands") ))

let initstate() = 
    svuota(continuation); svuota(tempstack)

	   
let loop (rho) = 
    
    while not(empty(continuation)) do
         let currconstr = top(continuation) in
	 (match currconstr with
	    | Expr1(e) -> itsem(rho)
	    | Expr2(e) -> itsem(rho)
	    | Exprd1(e) -> itsemden(rho)
	    | Exprd2(e) -> itsemden(rho)
	    | Coml(cl) -> itsemcl(rho)
	    | _ -> failwith("non legal construct in loop"))  
    done 
     
    
let sem (e,(r: dval env), (s: mval store)) = 
   initstate();
   globalstore := s;
   push(Expr1(e), continuation);
   loop(r);
   let valore= top(tempstack) in
   pop(tempstack); 
   valore
   

let semden (e,(r: dval env), (s: mval store)) = 
   initstate();
   globalstore := s;
   push(Exprd1(e), continuation);
   loop(r);
   let valore= top(tempdstack) in
   pop(tempdstack); 
   valore	  


   
let semcl (cl,(r: dval env), (s: mval store)) = 
   initstate();
   globalstore := s;
   push(labelcom(cl), continuation);
   loop(r);
   !globalstore 


