module Sem

(*SEMANTIC EVALUATION FUNCTIONS  *)

let rec sem ((e:exp), (r:dval env), (s: mval store)) =
  match e with
    | Eint(n) -> Int(n)
    | Ebool(b) -> Bool(b)
    | Den(i) -> dvaltoeval(applyenv(r,i))
    | Iszero(a) -> iszero(sem(a, r, s))
    | Eq(a,b) -> equ(sem(a, r, s), sem(b, r, s))
    | Prod(a,b) ->  mult (sem(a, r, s), sem(b, r, s))
    | Sum(a,b) ->  plus (sem(a, r, s), sem(b, r, s))
    | Diff(a,b)  ->  diff (sem(a, r, s), sem(b, r, s))
    | Minus(a) ->  minus(sem(a, r, s))
    | And(a,b) ->  et (sem(a, r, s), sem(b, r, s))
    | Or(a,b) ->  vel (sem(a, r, s), sem(b, r, s))
    | Not(a) -> non(sem(a, r, s))
    | Ifthenelse(a,b,c) ->
        let g = sem(a, r, s)
        if typecheck("bool",g) then
          (if g = Bool(true) then
             sem(b, r, s)
           else
             sem(c, r, s))
        else
          failwith ("nonboolean guard")
      | Val(e) ->
          match semden(e, r, s) with
            | Dloc n -> mvaltoeval(applystore(s, n))
            | _ -> failwith("not a variable")

and semden ((e:exp), (r:dval env), (s: mval store)) =
  match e with
  | Den(i) -> applyenv(r,i)
  | _ -> evaltodval(sem(e, r, s))

and semc((c: com), (r:dval env), (s: mval store)) =
  match c with
    | Assign(e1, e2) ->
        match semden(e1, r, s) with
          | Dloc(n) -> update(s, n, evaltomval(sem(e2, r, s)))
          | _ -> failwith ("wrong location in assignment")
    | Cifthenelse(e, cl1, cl2) ->
        let g = sem(e, r, s)
        if typecheck("bool",g) then
          (if g = Bool(true) then
             semcl(cl1, r, s)
           else
             semcl (cl2, r, s))
        else
          failwith ("nonboolean guard")
(* eliminato il punto fisso, utilizzando la sintassi, non composizionale *)
    | While(e, cl) ->
        let g = sem(e, r, s)
        if typecheck("bool",g) then
          (if g = Bool(true) then
             semcl((cl @ [While(e, cl)]), r, s)
           else
             s)
        else
          failwith ("nonboolean guard")

and semcl(cl, r, s) =
  match cl with
    | [] -> s
    | c::cl1 -> semcl(cl1, r, semc(c, r, s))









































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































































