module DominiStatic

(* tipo unico, diverso dal corrispettivo .ml*)
type evalenv = int

and eval =
     | Int of int
     | Bool of bool
     | Unbound
     | Funval of efun

(*adesso efun usa staticexp*)
and efun = staticexp * (evalenv)       


let (currentenv: evalenv ref) = ref(0)

let namestack = emptystack(stacksize,[| "dummy" |])

let evalstack = emptystack(stacksize,[| Unbound |])

let slinkstack = emptystack(stacksize, !currentenv)

type tag = Retained| Standard

let tagstack = emptystack(stacksize, Standard)


let emptyenv(x) =
  currentenv := -1
  svuota(namestack)
  svuota(evalstack)
  svuota(slinkstack)
  svuota(tagstack)
  !currentenv

let svuotaenv() =
  svuota(namestack)
  svuota(tagstack)
  svuota(evalstack)
  svuota(slinkstack)

let topenv() = !currentenv

let popenv () =
  pop(namestack)
  pop(evalstack)
  pop(slinkstack)
  pop(tagstack)
  currentenv := !currentenv - 1

let pushenv(r) =
  if r = !currentenv then
    push([||],namestack);
    push([||],evalstack);
    push(Standard,tagstack);
    push(r,slinkstack);
    currentenv := lungh(namestack)
  else
    currentenv := r


let retained (n:evalenv) = access(tagstack,n) = Retained

let retain () =
  setta(tagstack, !currentenv, Retained)
  let cont = ref(lungh(namestack))
  while !cont > -1 && retained(!cont) do
    cont := !cont - 1
  done
  currentenv := !cont


let collectretained (r) =
  let index = ref(lungh(namestack))
  while !index > r do
    pop(namestack)
    pop(evalstack)
    pop(slinkstack)
    pop(tagstack)
    index := !index - 1
  done
  currentenv := lungh(namestack)


(* ----------------StaticSem------------------- *)
(* utilizzato dall'analisi statica 

Notare che l'analizzatore StaticSem.exp lavora su exp e restituisce staticexp, mentre l'interprete StaticExpSem prende in input
staticexp e calcola eval, e' stato quindi necessario duplicare alcune funzioni per rendere compatibili i tipi (o alternativamente
avremmo potuto utilizzare funzioni polimorfe).
*)

type labeledconstruct =
   | Expr1 of exp
   | Expr2 of exp   

let (cstack: labeledconstruct stack stack) =
     emptystack(stacksize,emptystack(1,Expr1(Eint(0))))


let (tempsexpstack: staticexp stack stack) =
     emptystack(stacksize,emptystack(1,SUnbound))

(* ANALISI STATICA
    Questa e' la funzione che calcola gli offset, che vengono memorizzati in Access(!accesses,!index)
*)
let applyenv ((x: evalenv), (y: ide)) =
  let n = ref(x)
  let found = ref(false)
  let accesses = ref 0
  let index = ref 0
  while not !found && !n > -1 do
    let lenv = access(namestack,!n)
    let nl = Array.length lenv
    index := 0
    while not !found && !index < nl do
      if Array.get lenv !index = y then
        found := true
      else
        index := !index + 1
    done
    if not !found then
      n := access(slinkstack,!n)
      accesses := !accesses + 1
  done
  if !found then
    Access(!accesses, !index)
  else
    SUnbound

let bind ((r:evalenv),i,d) =
        push(Array.create 1 i,namestack);
        push(Array.create 1 Unbound,evalstack);
        push(r,slinkstack);
        push(Standard,tagstack);
        (lungh(namestack): evalenv)

let bindlist(r, il, el) =
   let n = List.length il in
   let ii = Array.create n "dummy" in
   let dd = Array.create n Unbound in
   let ri = ref(il) in
   let rd = ref(el) in
   let index = ref(0) in
   while !index < n do
     let i = List.head !ri in
         let d = List.head !rd in
         ( ri := List.tail !ri;
           rd := List.tail !rd;
           Array.set ii !index i;
           Array.set dd !index Unbound;
           index := !index + 1)
    done;
    push(ii, namestack);
    push(dd,evalstack);
    push(r,slinkstack);
    push(Standard,tagstack);
    (lungh(namestack): evalenv)


let pushargs ((b: exp list),continuation) =
  let br = ref(b)
  while not(!br = []) do
    push(Expr1(List.head !br),continuation)
    br := List.tail !br
  done

let getargs ((b: exp list),(tempstack: staticexp stack)) =
  let br = ref(b)
  let er = ref([])
  while not(!br = []) do
    let arg=top(tempstack)
    pop(tempstack)
    er := !er @ [arg];
    br := List.tail !br
  done
  !er

let rec to_be_retained (v:eval) =
  match v with
    | Funval (e, r1) -> not(r1 < !currentenv)
    | _ -> false



let retainorpopenv (r, valore) =
   if (topenv() = r) then
     ()
   else
     if !currentenv < lungh(namestack) then
       retain()
     else
       if to_be_retained(valore) then
         retain()
       else
         popenv()
         let index = ref(!currentenv)
         while !index > r do
           if retained(!index) then
             currentenv := !currentenv - 1
             index := !index - 1
           else
             index := r
         done


(* ----------------StaticExpSem------------------- *)
(* utilizzato dall'interprete *)

(*
    Il nuovo interprete usera' una pila di staticexp invece di exp, in quanto dovra' valutare le espressioni statiche calcolate 
    da StaticSem.fs
*)
type labeledstaticconstruct =
   | SExpr1 of staticexp
   | SExpr2 of staticexp


let (scstack: labeledstaticconstruct stack stack) =
     emptystack(stacksize,emptystack(1,SExpr1(SUnbound)))


let (tempvalstack: eval stack stack) =
     emptystack(stacksize,emptystack(1,Unbound))

(*l'accesso all'ambiente e' diretto dato che abbiamo gia' calcolato gli offset*)
let accessenv ((x: evalenv), (l: int), (index: int)) =
  let l = ref(l)
  let n = ref(x)
  while !l > 0 do
    n := access(slinkstack,!n)
    l := !l - 1
  Array.get (access(evalstack,!n)) index


let bind2 ((r:evalenv),d) =
        push(Array.create 1 "dummy",namestack);
        push(Array.create 1 d,evalstack);
        push(r,slinkstack);
        push(Standard,tagstack);
        (lungh(namestack): evalenv)



let bindlist2(r, el) =
   let n = List.length el in
   let ii = Array.create n "dummy" in
   let dd = Array.create n Unbound in
   let rd = ref(el) in
   let index = ref(0) in
   while !index < n do
         let d = List.head !rd in
         ( rd := List.tail !rd;
           Array.set dd !index d;
           index := !index + 1)
    done;
    push(ii,namestack);
    push(dd,evalstack);
    push(r,slinkstack);
    push(Standard,tagstack);
    (lungh(namestack): evalenv)

let pushargs2 ((b: staticexp list),continuation) =
  let br = ref(b)
  while not(!br = []) do
    push(SExpr1(List.head !br),continuation)
    br := List.tail !br
  done

let getargs2 ((b: staticexp list),(tempstack: eval stack)) =
  let br = ref(b)
  let er = ref([])
  while not(!br = []) do
    let arg=top(tempstack)
    pop(tempstack)
    er := !er @ [arg];
    br := List.tail !br
  done
  !er





