module StaticExpSem

(* SEMANTIC EVALUATION FUNCTIONS  *)

(*il nuovo interprete valutera' staticexp invece di exp*)

let newframes(e,rho) =
  let cframe = emptystack(cframesize(e),SExpr1(e))
  let tframe = emptystack(tframesize(e),Unbound)
  push(SExpr1(e),cframe)
  push(cframe,scstack)
  push(tframe,tempvalstack)
  pushenv(rho)

let makefun ((a:staticexp),(x:evalenv)) =
  match a with
    | SFun(aa) -> Funval(a,x)
    | _ -> failwith ("Non-functional object")

let applyfun ((ev1:eval),(ev2:eval list)) =
  match ev1 with
    | Funval(SFun(aa),r) -> newframes(aa,bindlist2(r, ev2))
    | _ -> failwith ("attempt to apply a non-functional object")


(* diversa: rfix diventa un puntatore *)
let makefunrec ((a:staticexp),(x:evalenv)) =
  makefun(a, ((lungh(namestack) + 1): evalenv))

let sem ((e:staticexp), (r:evalenv)) =
  push(emptystack(1,Unbound),tempvalstack)
  newframes(e,r)
  while not(empty(scstack)) do
    while not(empty(top(scstack))) do
      let continuation = top(scstack)
      let tempstack = top(tempvalstack)
      let rho = topenv()
      match top(continuation) with
        | SExpr1(x) ->
            pop(continuation)
            push(SExpr2(x),continuation)
            match x with
              | Siszero(a) -> push(SExpr1(a),continuation)
              | Sequ(a,b) ->
                  push(SExpr1(a),continuation)
                  push(SExpr1(b),continuation)
              | Smult(a,b) ->
                push(SExpr1(a),continuation)
                push(SExpr1(b),continuation)
              | Splus(a,b) ->
                push(SExpr1(a),continuation)
                push(SExpr1(b),continuation)
              | Sdiff(a,b) ->
                  push(SExpr1(a),continuation)
                  push(SExpr1(b),continuation)
              | Sminus(a) -> push(SExpr1(a),continuation)
              | Set(a,b) ->
                push(SExpr1(a),continuation)
                push(SExpr1(b),continuation)
              | Svel(a,b) ->
                  push(SExpr1(a),continuation)
                  push(SExpr1(b),continuation)
              | Snon(a) -> push(SExpr1(a),continuation)
              | Sifthenelse(a,b,c) -> push(SExpr1(a),continuation)
              | SBind(e1,e2) -> push(SExpr1(e1),continuation)
              | SAppl(a,b) ->
                  push(SExpr1(a),continuation)
                  pushargs2(b,continuation)
              | _ -> ()
        | SExpr2(x) ->
            pop(continuation)
            match x with
              | SUnbound -> push(Unbound,tempstack)
              | SInt(n) ->  push(Int(n),tempstack)
              | SBool(b) -> push(Bool(b),tempstack)
              | Access(l,i) ->   push(accessenv(rho,l,i),tempstack) (*accesso diretto all'ambiente*)
              | Siszero(a) ->
                  let arg=top(tempstack)
                  pop(tempstack)
                  push(iszero(arg),tempstack)
              | Sequ(a,b) ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=top(tempstack)
                  pop(tempstack)
                  push(equ(firstarg,sndarg),tempstack)
              | Smult(a,b) ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=top(tempstack)
                  pop(tempstack)
                  push(mult(firstarg,sndarg),tempstack)
              | Splus(a,b) ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=top(tempstack)
                  pop(tempstack)
                  push(plus(firstarg,sndarg),tempstack)
              | Sdiff(a,b)  ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=top(tempstack)
                  pop(tempstack)
                  push(diff(firstarg,sndarg),tempstack)
              | Sminus(a) ->
                  let arg=top(tempstack)
                  pop(tempstack)
                  push(minus(arg),tempstack)
              | Set(a,b) ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=top(tempstack)
                  pop(tempstack)
                  push(et(firstarg,sndarg),tempstack)
              | Svel(a,b) ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=top(tempstack)
                  pop(tempstack)
                  push(vel(firstarg,sndarg),tempstack)
              | Snon(a) ->
                  let arg=top(tempstack)
                  pop(tempstack)
                  push(non(arg),tempstack)
              | Sifthenelse(a,b,c) ->
                  let arg=top(tempstack)
                  pop(tempstack)
                  match arg with
                    | Bool(bg) ->
                        if bg  then
                          push(SExpr1(b),continuation)
                        else
                          push(SExpr1(c),continuation)
                    | _ -> failwith ("type error")
              | SFun(a) ->  push(makefun(SFun(a),rho),tempstack)
              | SRec(e) -> push(makefunrec(e,rho),tempstack)
              | SBind(e1,e2) ->
                  let arg=top(tempstack)
                  pop(tempstack)
                  newframes(e2,bind2(rho, arg))
              | SAppl(a,b) ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=getargs2(b,tempstack)
                  applyfun(firstarg,sndarg)
    done
    let valore= top(top(tempvalstack))
    pop(top(tempvalstack))
    pop(scstack)
    pop(tempvalstack)
    push(valore,top(tempvalstack))
    retainorpopenv (r, valore) (* cambiato per la retention *)
  done
  collectretained(r)  (* cambiato per la retention *)
  let valore = top(top(tempvalstack))
  pop(tempvalstack)
  valore
