module StaticSem
open StaticSyntax
open ImpMPila
open DominiStatic
open Soperations

(*ANALISI STATICA: TRADUCIAMO TUTTI I DEN IDE IN ACCESSI DIRETTI CALCOLANDO L'OFFSET *)

let newframes(e,rho) =
  let cframe = emptystack(cframesize(e),Expr1(e))
  let tframe = emptystack(tframesize(e),SUnbound)
  push(Expr1(e),cframe)
  push(cframe,cstack)
  push(tframe,tempsexpstack)
  pushenv(rho)

let sem ((e:exp), (r:evalenv)) =
  push(emptystack(1,SUnbound),tempsexpstack)
  newframes(e,r)
  while not(empty(cstack)) do
    while not(empty(top(cstack))) do
      let continuation = top(cstack)
      let tempstack = top(tempsexpstack)
      let rho = topenv()
      match top(continuation) with
        | Expr1(x) ->
            pop(continuation)
            push(Expr2(x),continuation)
            match x with
              | Iszero(a) -> push(Expr1(a),continuation)
              | Eq(a,b) ->
                  push(Expr1(a),continuation)
                  push(Expr1(b),continuation)
              | Prod(a,b) ->
                push(Expr1(a),continuation)
                push(Expr1(b),continuation)
              | Sum(a,b) ->
                push(Expr1(a),continuation)
                push(Expr1(b),continuation)
              | Diff(a,b) ->
                  push(Expr1(a),continuation)
                  push(Expr1(b),continuation)
              | Minus(a) -> push(Expr1(a),continuation)
              | And(a,b) ->
                push(Expr1(a),continuation)
                push(Expr1(b),continuation)
              | Or(a,b) ->
                  push(Expr1(a),continuation)
                  push(Expr1(b),continuation)
              | Not(a) -> push(Expr1(a),continuation)
              | Ifthenelse(a,b,c) ->        (*staticamente eseguiamo entrambi i rami dell'if per tradurre i Den(i) interni*)
                  push(Expr1(a),continuation)
                  push(Expr1(b),continuation)
                  push(Expr1(c),continuation)
              | Let(i,e1,e2) -> push(Expr1(e1),continuation)   (* Prima di chiudere la funzione in una espressione statica bisogna
                                                                               valutare il corpo in un nuovo frame*)
                                newframes(e2,bind(rho, i, SUnbound))
              | Fun(i,a) ->
                  let foo = List.init (List.length i) (fun _ -> Unbound)    (* ignoriamo i valori dei parametri, usiamo una lista dummy*)
                  newframes(a,bindlist(rho, i, foo))                        (* prima di chiudere la funzione in una espressione statica bisogna
                                                                           valutare il corpo in un nuovo frame*)
              | Rec(f,e) ->
                  newframes(e,bind(rho, f, Unbound))                     (*  i costrutti Rec sono sempre nella forma 
                                                                                 Let ( ide, (Rec ( ide, Fun(... ) ), ... )
                                                                                 Bisogna creare un nuovo frame con il binding della funzione ricorsiva,
                                                                                 Successivamente il costrutto Fun costruira' un nuovo frame contenente i
                                                                                 parametri della funzione.
                                                                                 Prima di fare la chiusura per il costrutto Rec (verra' fatto nel ramo 
                                                                                  Exp2) bisogna valutare il corpo per effuttare eventuali traduzioni.
                                                                             *)
              | Appl(a,b) ->
                  push(Expr1(a),continuation)
                  pushargs(b,continuation)
              | _ -> ()
        | Expr2(x) ->
            pop(continuation)
            match x with
              | Eint(n) ->  push(SInt(n),tempstack)
              | Ebool(b) -> push(SBool(b),tempstack)
              | Den(i) ->   push(applyenv(rho,i),tempstack) (* qui verra' tradotto in Access(int,int)*)
              | Iszero(a) ->
                  let arg=top(tempstack)
                  pop(tempstack)
                  push(Siszero(arg),tempstack)
              | Eq(a,b) ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=top(tempstack)
                  pop(tempstack)
                  push(Sequ(firstarg,sndarg),tempstack)
              | Prod(a,b) ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=top(tempstack)
                  pop(tempstack)
                  push(Smult(firstarg,sndarg),tempstack)
              | Sum(a,b) ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=top(tempstack)
                  pop(tempstack)
                  push(Splus(firstarg,sndarg),tempstack)
              | Diff(a,b)  ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=top(tempstack)
                  pop(tempstack)
                  push(Sdiff(firstarg,sndarg),tempstack)
              | Minus(a) ->
                  let arg=top(tempstack)
                  pop(tempstack)
                  push(Sminus(arg),tempstack)
              | And(a,b) ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=top(tempstack)
                  pop(tempstack)
                  push(Set(firstarg,sndarg),tempstack)
              | Or(a,b) ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=top(tempstack)
                  pop(tempstack)
                  push(Svel(firstarg,sndarg),tempstack)
              | Not(a) ->
                  let arg=top(tempstack)
                  pop(tempstack)
                  push(Snon(arg),tempstack)
              | Ifthenelse(a,b,c) ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=top(tempstack)
                  pop(tempstack)
                  let trdarg=top(tempstack)
                  pop(tempstack)
                  push(Sifthenelse(firstarg,sndarg,trdarg),tempstack)
              | Fun(i,a) ->
                  let arg=top(tempstack)
                  pop(tempstack)
                  push(SFun(arg),tempstack)
              | Rec(f,e) ->
                  let arg=top(tempstack)
                  pop(tempstack)
                  push(SRec(arg),tempstack)                                                                                                  
              | Let(i,e1,e2) ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=top(tempstack)
                  pop(tempstack)
                  push(SBind(firstarg, sndarg),tempstack)
              | Appl(a,b) ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=getargs(b,tempstack)
                  push(SAppl(firstarg,sndarg),tempstack)
    done
    let valore= top(top(tempsexpstack))
    pop(top(tempsexpstack))
    pop(cstack)
    pop(tempsexpstack)
    push(valore,top(tempsexpstack))
    popenv()
    (*retainorpopenv (r, valore)*) (* cambiato per la retention *)
  done
  (*collectretained(r)*)  (* cambiato per la retention *)
  let valore = top(top(tempsexpstack))
  pop(tempsexpstack)
  valore
