module Sem

(*SEMANTIC EVALUATION FUNCTIONS  *)

let newframes(e,rho) =
  let cframe = emptystack(cframesize(e),Expr1(e))
  let tframe = emptystack(tframesize(e),Unbound)
  push(Expr1(e),cframe)
  push(cframe,cstack)
  push(tframe,tempvalstack)
  pushenv(rho)

let makefun ((a:exp),(x:eval env)) =
  match a with
    | Fun(ii,aa) -> Funval(a,x)
    | _ -> failwith ("Non-functional object")

let applyfun ((ev1:eval),(ev2:eval list)) =
  match ev1 with
    | Funval(Fun(ii,aa),r) -> newframes(aa,bindlist(r, ii, ev2))
    | _ -> failwith ("attempt to apply a non-functional object")

let makefunrec (i,e1,r) =
  let functional rr =
    bind(r, i, makefun(e1,rr))
  let rec rfix = function x -> functional rfix x
  makefun(e1, rfix)

let sem ((e:exp), (r:eval env)) =
  push(emptystack(1,Unbound),tempvalstack)
  newframes(e,r)
  while not(empty(cstack)) do
    while not(empty(top(cstack))) do
      let continuation = top(cstack)
      let tempstack = top(tempvalstack)
      let rho = topenv() in
      match top(continuation) with
        | Expr1(x) ->
            pop(continuation)
            push(Expr2(x),continuation)
            match x with
              | Iszero(a) -> push(Expr1(a),continuation)
              | Eq(a,b) ->
                  push(Expr1(a),continuation)
                  push(Expr1(b),continuation)
              | Prod(a,b) ->
                push(Expr1(a),continuation)
                push(Expr1(b),continuation)
              | Sum(a,b) ->
                  push(Expr1(a),continuation)
                  push(Expr1(b),continuation)
              | Diff(a,b) ->
                  push(Expr1(a),continuation)
                  push(Expr1(b),continuation)
              | Minus(a) -> push(Expr1(a),continuation)
              | And(a,b) ->
                  push(Expr1(a),continuation)
                  push(Expr1(b),continuation)
              | Or(a,b) ->
                  push(Expr1(a),continuation)
                  push(Expr1(b),continuation)
              | Not(a) -> push(Expr1(a),continuation)
              | Ifthenelse(a,b,c) -> push(Expr1(a),continuation)
              | Let(i,e1,e2) -> push(Expr1(e1),continuation)
              | Appl(a,b) ->
                  push(Expr1(a),continuation)
                  pushargs(b,continuation)
              | _ -> ()
        | Expr2(x) ->
            pop(continuation)
            match x with
              | Eint(n) ->  push(Int(n),tempstack)
              | Ebool(b) -> push(Bool(b),tempstack)
              | Den(i) ->   push(applyenv(rho,i),tempstack)
              | Iszero(a) ->
                  let arg=top(tempstack)
                  pop(tempstack)
                  push(iszero(arg),tempstack)
              | Eq(a,b) ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=top(tempstack)
                  pop(tempstack)
                  push(equ(firstarg,sndarg),tempstack)
              | Prod(a,b) ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=top(tempstack)
                  pop(tempstack)
                  push(mult(firstarg,sndarg),tempstack)
              | Sum(a,b) ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=top(tempstack)
                  pop(tempstack)
                  push(plus(firstarg,sndarg),tempstack)
              | Diff(a,b)  ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=top(tempstack)
                  pop(tempstack)
                  push(diff(firstarg,sndarg),tempstack)
              | Minus(a) ->
                  let arg=top(tempstack)
                  pop(tempstack)
                  push(minus(arg),tempstack)
              | And(a,b) ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=top(tempstack)
                  pop(tempstack)
                  push(et(firstarg,sndarg),tempstack)
              | Or(a,b) ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=top(tempstack)
                  pop(tempstack)
                  push(vel(firstarg,sndarg),tempstack)
              | Not(a) ->
                  let arg=top(tempstack)
                  pop(tempstack)
                  push(non(arg),tempstack)
              | Ifthenelse(a,b,c) ->
                  let arg=top(tempstack)
                  pop(tempstack)
                  match arg with
                    | Bool(bg) ->
                      if bg then
                        push(Expr1(b),continuation)
                      else
                        push(Expr1(c),continuation)
                    | _ -> failwith ("type error")
              | Fun(i,a) ->  push(makefun(Fun(i,a),rho),tempstack)
              | Rec(f,e) -> push(makefunrec(f,e,rho),tempstack)
              | Let(i,e1,e2) ->
                  let arg=top(tempstack)
                  pop(tempstack)
                  newframes(e2,bind(rho, i, arg))
              | Appl(a,b) ->
                  let firstarg=top(tempstack)
                  pop(tempstack)
                  let sndarg=getargs(b,tempstack)
                  applyfun(firstarg,sndarg)
    done
    let valore= top(top(tempvalstack))
    pop(top(tempvalstack))
    popenv()
    pop(cstack)
    pop(tempvalstack)
    push(valore,top(tempvalstack))
  done
  let valore = top(top(tempvalstack))
  pop(top(tempvalstack))
  pop(tempvalstack)
  valore
