type 't env = int
type loc = int * int
type 't store = int 

exception Nonstorable
exception Nonexpressible

type eval =  	   
	| Int of int 
	| Bool of bool
	| Novalue
	| Funval of efun  
and efun =  exp * (dval env)
and proc =  exp * (dval env)

		
and dval =  	   
	| Dint of int 
	| Dbool of bool 
	| Unbound
		| Dloc of loc
	| Dfunval of efun
		| Dprocval of proc 
	
and mval = 
	| Mint of int
		| Mbool of bool
		| Undefined
	
let evaltomval e = match e with
	| Int n -> Mint n
		| Bool n -> Mbool n
		| _ -> raise Nonstorable
	
let mvaltoeval m = match m with
	| Mint n -> Int n
		| Mbool n -> Bool n
		| _ -> Novalue
	
let evaltodval e = match e with
	| Int n -> Dint n
		| Bool n -> Dbool n
		| Novalue -> Unbound
		| Funval f -> Dfunval f
	
let dvaltoeval e = match e with
	| Dint n -> Int n
		| Dbool n -> Bool n
		| Dloc n -> raise Nonexpressible
		| Dprocval n -> raise Nonexpressible
		| Unbound -> Novalue
		| Dfunval f -> Funval f


type labeledconstruct = 
   | Expr1 of exp 
   | Expr2 of exp
   | Exprd1 of exp
   | Exprd2 of exp
   | Com1 of com
   | Com2 of com
   | Coml of labeledconstruct list
   | Decl of labeledconstruct list
   | Rdecl of (ide * exp) list
   | Dec1 of (ide * exp) 
   | Dec2 of (ide * exp) 

	 
let (cstack: labeledconstruct stack stack) = 
  emptystack(stacksize,emptystack(1,Expr1(Eint(0))))

let (tempvalstack: eval stack stack) = emptystack(stacksize,emptystack(1,Novalue))

let (tempdvalstack: dval stack stack) = emptystack(stacksize,emptystack(1,Unbound))

let labelcom (dl: com list) = let dlr = ref(dl) in
	let ldlr = ref([]) in
	while not (!dlr = []) do
	 let i = List.hd !dlr in
	 ldlr := !ldlr @ [Com1(i)];
	 dlr := List.tl !dlr
	done;
	Coml(!ldlr)
	
let labeldec (dl: (ide * exp) list) = let dlr = ref(dl) in
	let ldlr = ref([]) in
	while not (!dlr = []) do
	 let i = List.hd !dlr in
	 ldlr := !ldlr @ [Dec1(i)];
	 dlr := List.tl !dlr
	done;
	Decl(!ldlr)
	 
let (labelstack: labeledconstruct stack) =
	  emptystack(stacksize,Expr1(Eint(0))) 


let pushargs ((b: exp list),(continuation: labeledconstruct stack)) = 
     let br = ref(b) in
      while not(!br = []) do
	    push(Exprd1(List.hd !br),continuation);
	    br := List.tl !br
	  done

let getargs ((b: exp list),(tempstack: dval stack)) = 
      let br = ref(b) in
      let er = ref([]) in
      while not(!br = []) do
	    let arg=top(tempstack) in
	pop(tempstack); er := !er @ [arg];
	    br := List.tl !br
	  done; 
	  !er

	  (* ambiente *)
let (currentenv: dval env ref) = ref(0)
 
let (initenv: dval env ref) = ref(0) (* nuovo *)

let namestack = emptystack(stacksize,[| "dummy" |])

let dvalstack = emptystack(stacksize,[| Unbound |])

let slinkstack = emptystack(stacksize, !currentenv)

type tag = Retained| Standard

let tagstack = emptystack(stacksize, Standard)

let applyenv ((x: dval env), (y: ide)) = 
       let n = ref(x) in
       let den = ref(Unbound) in
       while !n > -1 do
	     let lenv = access(namestack,!n) in
		 let nl = Array.length lenv in
		 let index = ref(0) in
		 while !index < nl do
	  if Array.get lenv !index = y then
	  (den := Array.get (access(dvalstack,!n)) !index;
	  index := nl)
		  else index := !index + 1
		 done;
	 if not(!den = Unbound) then  n := -1
		 else n := access(slinkstack,!n)
       done;
       !den
	  


let bind ((r:dval env),i,d) = 
	let arrnomi = access(namestack, r) in
	let lun =  Array.length arrnomi in
	let arrden = access(dvalstack, r) in
	let n = ref(0) in
	let index = ref(-1) in
	while !n < lun do
	 let nome = Array.get arrnomi !n in 
	 if nome = i then 
	   (Array.set arrden !n d;
	   n := lun; index := -2) else  
	 if nome = "dummy" 
	 then (index := !n; n := lun)
	 else n := !n + 1
	done;
	if !index = -2 then r else 
	if !index > -1 
	then
	   (Array.set arrnomi !index i;
	   Array.set arrden !index d; r)
	else 
	   (push(Array.create 1 i,namestack); 
	   push(Array.create 1 d,dvalstack);
	   push(r,slinkstack);
	   push(Standard,tagstack);
	   (lungh(namestack): dval env))
	   
let bindlist(r, il, el) = 
   let n = List.length il in
   let ii = Array.create n "dummy" in
   let dd = Array.create n Unbound in
   let ri = ref(il) in
   let rd = ref(el) in
   let index = ref(0) in
   while !index < n do
     let i = List.hd !ri in
	 let d = List.hd !rd in
	 ( ri := List.tl !ri;
	   rd := List.tl !rd;
	   Array.set ii !index i;
	   Array.set dd !index d;
	   index := !index + 1)
    done;
    push(ii, namestack);
    push(dd,dvalstack);
    push(r,slinkstack);
    push(Standard,tagstack);
    (lungh(namestack): dval env)
	
let emptyenv(x) = currentenv := -1; svuota(namestack);
   svuota(dvalstack); svuota(slinkstack); svuota(tagstack);
   !currentenv

let svuotaenv() = svuota(namestack); svuota(tagstack);
    svuota(dvalstack); svuota(slinkstack)


(* nuova per i blocchi *)
let pushlocalenv (dl,dlr,r) = 
   let anomi = access(namestack, r) in
   let aden = access(dvalstack, r) in
   let nn = ref(Array.length anomi) in
   let rn = (List.length (dl @ dlr) + !nn) in
   let ii = Array.create rn "dummy" in
   let dd = Array.create rn Unbound in
   while not(!nn = 0) do
      Array.set ii (!nn - 1) (Array.get anomi (!nn - 1));
      Array.set dd (!nn - 1) (Array.get aden (!nn - 1));
      nn := !nn - 1
   done;
   pop(namestack); pop(dvalstack);
   push(ii, namestack);
   push(dd, dvalstack);
   r


let topenv() = !currentenv

let popenv () = pop(namestack);
	  pop(dvalstack);
	  pop(slinkstack);
	  pop(tagstack);
	  currentenv := !currentenv - 1
	  
let pushemptylocalenv(r) = push([||],namestack);
	   push([||],dvalstack);
	   push(Standard,tagstack);
	   push(r,slinkstack);
	   currentenv := lungh(namestack)
	   
let pushenv(r) = if r = !currentenv then 
	   pushemptylocalenv(r)
	 else currentenv := r
	  
(* retention *)
let retained (n:dval env) = access(tagstack,n) = Retained 
				

let retain () = 
       setta(tagstack, !currentenv, Retained);
       let cont = ref(lungh(namestack)) in
       while !cont > -1  & retained(!cont) do cont := !cont - 1 done;
       currentenv := !cont
       
let to_be_retained (v:eval) = match v with
   | Funval (e, r1) -> not(r1 < !currentenv)
   | _ -> false

let retainorpopenv (r, valore) =
   if (topenv() = r) then () else
   if !currentenv < lungh(namestack) then retain() else
   if to_be_retained(valore) then retain() else
      (popenv();
      let index = ref(!currentenv) in
	  while !index > r do
	   if retained(!index) then (currentenv := !currentenv - 1;
	   index := !index -1) else (index := r) done )
	   

   (* implementazione store  *) 

let (currentstore: mval store ref) = ref(0)

let storestack = emptystack(stacksize,[|Undefined|])


let (newloc,initloc) =
   let count = ref(-1) in
   (fun () -> count := !count +1;
	      (!currentstore, !count)),
   (fun () -> count := -1)
   
let applystore ((x: mval store), ((n1, n2): loc)) = 
       let a = access(storestack, n1) in
	    Array.get a n2
	   
let emptystore(x) = initloc(); 
   svuota(storestack); currentstore := -1; !currentstore 

let allocate ((s:mval store), (m:mval)) = let (n1, n2) = newloc() in 
    let a = access(storestack, n1) in
		   Array.set a n2 m; ((n1, n2), s)
	
let update((s:mval store),  (n1,n2),  (m:mval)) = 
     if applystore(s, (n1,n2)) = Undefined 
	 then failwith ("wrong assignment") 
	 else let a = access(storestack, n1) in
		   Array.set a n2 m; s
	   
let pushlocalstore (dl) = let rdl = ref(dl) in
   let rn = ref(0) in
   while not(!rdl = []) do
      let (i, d) = List.hd !rdl in
	  (match d with
	    | Newloc(_) -> rn := !rn + 1
	    | _ -> ());
      rdl := List.tl !rdl
   done;
   let a = Array.create !rn Undefined in
   pop(storestack); push(a, storestack); 
   initloc(); !currentstore
   

let pushemptylocalstore () = push([||],storestack); 
   currentstore := !currentstore + 1

let pushstore(n) = if n = !currentstore 
        then pushemptylocalstore()
	else currentstore := n

let popstore () = pop(storestack); currentstore := !currentstore -1

let svuotastore () = svuota(storestack)

let topstore() = !currentstore;;


