
(*SEMANTIC EVALUATION FUNCTIONS  *)

let newframes(ss,rho,sigma) =
 pushstore(sigma);
 pushenv(rho);
 let cframe = emptystack(cframesize(ss),Expr1(Eint 0)) in
 let tframe = emptystack(tframesize(ss),Novalue) in
 let dframe = emptystack(tdframesize(ss),Unbound) in
 push(tframe,tempvalstack);   
 push(dframe,tempdvalstack);
 push(ss, labelstack);
 push(ss, cframe);
 push(cframe, cstack)
 
	

let makefun ((a:exp),(x:dval env)) =
       (match a with
       |	Fun(ii,aa) ->
	   Dfunval(a,x)
       |	_ -> failwith ("Non-functional object"))

let makeproc ((a:exp),(x:dval env)) =
     (match a with
       |	Proc(ii,b) ->
	   Dprocval(a,x) 
       |	_ -> failwith ("Non-procedural object"))
       
(* diversa *)      
let makefunrec (f, (a:exp),(x:dval env)) =
	     dvaltoeval(makefun(a, lungh(namestack) + 1))

let applyfun ((ev1:dval),(ev2:dval list), s) =
       ( match ev1 with
       | Dfunval(Fun(ii,aa),r) -> newframes(Expr1(aa),bindlist(r, ii, ev2), s)
       | _ -> failwith ("attempt to apply a non-functional object"))
       
(* cambiata *)
let applyproc ((ev1:dval),(ev2:dval list), s) =
       ( match ev1 with
       | Dprocval(Proc(ii,(l1, l2, l3)), x) -> 
             let r = bindlist(x, ii, ev2) in
	     newframes(labelcom(l3), r, s);
	     push(Rdecl(l2), top(cstack));
	     push(labeldec(l1),top(cstack));
	     pushlocalenv(l1,l2,r);
	     pushlocalstore(l1);
	     ()
       | _ -> failwith ("attempt to apply a non-functional object"))

let itsem() =
    let continuation = top(cstack) in
    let tempstack = top(tempvalstack) in
    let tempdstack = top(tempdvalstack) in
    let rho = topenv() in
    let sigma = topstore() in
    (match top(continuation) with
     |Expr1(x) -> 
	 (pop(continuation);
	 push(Expr2(x),continuation); 
	 (match x with
      
	  | Iszero(a) -> push(Expr1(a),continuation)
	  | Eq(a,b) -> push(Expr1(a),continuation); 
		push(Expr1(b),continuation)
	  | Prod(a,b) ->  push(Expr1(a),continuation); 
		push(Expr1(b),continuation)
	  | Sum(a,b) -> push(Expr1(a),continuation); 
		push(Expr1(b),continuation)
	  | Diff(a,b)  ->  push(Expr1(a),continuation); 
		push(Expr1(b),continuation)
	  | Minus(a) ->  push(Expr1(a),continuation)
	  | And(a,b) ->  push(Expr1(a),continuation); 
		push(Expr1(b),continuation)
	  | Or(a,b) ->  push(Expr1(a),continuation); 
		push(Expr1(b),continuation)
	  | Not(a) -> push(Expr1(a),continuation)
	  | Ifthenelse(a,b,c) -> push(Expr1(a),continuation)
	  | Val(e) ->  push(Exprd1(e),continuation)
	  | Newloc(e) -> failwith ("nonlegal expression for sem")
	  | Let(i,e1,e2) ->  push(Exprd1(e1),continuation)
	  | Appl(a,b) ->  push(Expr1(a),continuation); 
	       pushargs(b,continuation)
	  | Proc(i,b) -> failwith ("nonlegal expression for sem") 
	  | _ -> ()))
    |Expr2(x) -> 
	 (pop(continuation); 
	 (match x with
	  | Eint(n) ->  push(Int(n),tempstack)
	  | Ebool(b) -> push(Bool(b),tempstack)
	  | Den(i) ->   push(dvaltoeval(applyenv(rho,i)),tempstack)
	  | Iszero(a) -> 
	     let arg=top(tempstack) in
	     pop(tempstack);
	     push(iszero(arg),tempstack)
	  | Eq(a,b) -> 
	     let firstarg=top(tempstack) in
	     pop(tempstack);
	     let sndarg=top(tempstack) in
	     pop(tempstack);
	     push(equ(firstarg,sndarg),tempstack)
	  | Prod(a,b) -> 
	     let firstarg=top(tempstack) in
	     pop(tempstack);
	     let sndarg=top(tempstack) in
	     pop(tempstack);
	     push(mult(firstarg,sndarg),tempstack)  
	  | Sum(a,b) -> 
	     let firstarg=top(tempstack) in
	     pop(tempstack);
	     let sndarg=top(tempstack) in
	     pop(tempstack);
	     push(plus(firstarg,sndarg),tempstack)
	  | Diff(a,b)  -> 
	     let firstarg=top(tempstack) in
	     pop(tempstack);
	     let sndarg=top(tempstack) in
	     pop(tempstack);
	     push(diff(firstarg,sndarg),tempstack) 
	  | Minus(a) -> 
	     let arg=top(tempstack) in
	     pop(tempstack);
	     push(minus(arg),tempstack)  
	  | And(a,b) -> 
	     let firstarg=top(tempstack) in
	     pop(tempstack);
	     let sndarg=top(tempstack) in
	     pop(tempstack);
	     push(et(firstarg,sndarg),tempstack)  
	  | Or(a,b) -> 
	     let firstarg=top(tempstack) in
	     pop(tempstack);
	     let sndarg=top(tempstack) in
	     pop(tempstack);
	     push(vel(firstarg,sndarg),tempstack)  
	  | Not(a) -> 
	     let arg=top(tempstack) in
	     pop(tempstack);
	     push(non(arg),tempstack) 
	  | Ifthenelse(a,b,c) ->
	     let arg=top(tempstack) in
	     pop(tempstack);
	     if typecheck("bool",arg) then
	       (if arg = Bool(true) 
	       then push(Expr1(b),continuation)
	       else push(Expr1(c),continuation))
	    else failwith ("type error") 
	  | Val(e) -> let v = top(tempdstack) in
		 pop(tempdstack);
			 (match v with 
		       | Dloc n -> 
			   push(mvaltoeval(applystore(sigma, n)), tempstack)
			   | _ -> failwith("not a variable"))            
	  | Fun(i,a) ->  push(dvaltoeval(makefun(Fun(i,a),rho)),tempstack)
	  | Rec(f,e) -> push(makefunrec(f,e,rho),tempstack)
	  | Let(i,e1,e2) -> let arg= top(tempdstack) in
	     pop(tempdstack); newframes(Expr1(e2), bind(rho, i, arg), sigma)
	  | Appl(a,b) -> let firstarg=evaltodval(top(tempstack)) in
	     pop(tempstack); 
	     let sndarg=getargs(b,tempdstack) in
	     applyfun(firstarg, sndarg, sigma)
	 | _ -> failwith("no more cases for semexpr")))
		| _ -> failwith("no more cases for semexpr")) 
	  
let itsemden() =
    let continuation = top(cstack) in
    let tempstack = top(tempvalstack) in
    let tempdstack = top(tempdvalstack) in
    let rho = topenv() in
    let sigma = topstore() in 
    (match top(continuation) with
     |Exprd1(x) -> (pop(continuation); push(Exprd2(x),continuation);
       match x with
	  | Den i -> ()
	  | Fun(i,e) -> ()
	  | Proc(il,b) -> ()
	  | Newloc(e)  -> push(Expr1( e), continuation)
	  | _ -> push(Expr1(x), continuation))          
          |Exprd2(x) -> (pop(continuation); match x with
	   | Den i -> push(applyenv(rho,i), tempdstack)
	   | Fun(i,e)  -> push(makefun(x,rho),tempdstack)
	   | Proc(il,b)  -> push(makeproc(x,rho),tempdstack)
	   | Newloc(e)  ->let m=evaltomval(top(tempstack)) in
		pop(tempstack); let (l, s1)  = allocate(sigma, m) in  
				push(Dloc l, tempdstack);
				currentstore := s1 (* cambiata *)
	   | _ -> let arg = top(tempstack) in pop(tempstack);
	      push(evaltodval(arg), tempdstack))
     | _ -> failwith("No more cases for demden")    )
	 
let itsemdecl () =
    let tempstack = top(tempvalstack) in
    let continuation = top(cstack) in
    let tempdstack = top(tempdvalstack) in
    let rho = topenv() in
    let sigma = topstore() in
    let dl =
    (match top(continuation) with
      | Decl(dl1) -> dl1
      | _ -> failwith("impossible in semdecl")) in
    if dl = [] then pop(continuation) else
    (let currd = List.hd dl in
    let newdl = List.tl dl in
    pop(continuation); push(Decl(newdl),continuation);
    (match currd with
    
	| Dec1( (i,e)) -> 
	   pop(continuation); 
	   push(Decl(Dec2((i, e))::newdl),continuation);
	   push(Exprd1(e), continuation)
			 
	| Dec2((i,e)) -> 
	   let arg = top(tempdstack) in 
	   pop(tempdstack); 
	   currentenv := bind(rho, i, arg)   (* cambiata *)	 
	| _ -> failwith("no more sensible cases for semdecl")))	
	
	
let itsemrdecl() = 
	    let tempstack = top(tempvalstack) in
	    let continuation = top(cstack) in
	    let tempdstack = top(tempdvalstack) in
	    let rho = topenv() in
	    let sigma = topstore() in
	    let rl = (match top(continuation) with
	      | Rdecl(rl1) -> rl1 
	      | _ -> failwith("impossible in semrdecl")) in
	    pop(continuation);
	    (* tolto il punto fisso del funzionale *)
		   let prl = ref(rl) in
		   while not(!prl = []) do
		      let currd = List.hd !prl in
		      prl := List.tl !prl;
		      let (i, den) =
		      (match currd with
		       |(j,  Proc(il,b)) -> (j, makeproc(Proc(il,b),rho))
		       |(j,  Fun(il,b)) -> (j, makefun(Fun(il,b),rho))
		       | _ -> failwith("no more sensible cases in recursive declaration")) in
		      currentenv := bind(rho, i, den) 
		   done 
	       
let itsemcl() =
    let tempstack = top(tempvalstack) in
	let continuation = top(cstack) in
    let tempdstack = top(tempdvalstack) in
    let rho = topenv() in
    let sigma = topstore() in
	let cl =
    (match top(continuation) with
      | Coml(dl1) -> dl1
      | _ -> failwith("impossible in semcl")) in
    if cl = [] then pop(continuation) else
    (let currc = List.hd cl in
    let newcl = List.tl cl in
    pop(continuation); push(Coml(newcl),continuation);
	(match currc with
	   | Com1(Assign(e1, e2)) ->  pop(continuation); 
		push(Coml(Com2(Assign(e1, e2))::newcl),continuation);
		push(Exprd1(e1), continuation);
		push(Expr1(e2), continuation)		
	   | Com2(Assign(e1, e2)) -> 
	      let arg2 = evaltomval(top(tempstack)) in
	      pop(tempstack); 
	      let arg1 = top(tempdstack) in
	      pop(tempdstack); 
	      (match arg1 with
		  | Dloc(n) -> currentstore := update(sigma, n, arg2) (* cambiato *)				   
		  | _ -> failwith ("wrong location in assignment"))		  
	   | Com1(While(e, cl)) ->
		pop(continuation); 
		push(Coml(Com2(While(e, cl))::newcl),continuation);
		push(Expr1(e), continuation)	
	   | Com2(While(e, cl)) -> 
		 let g = top(tempstack) in
		 pop(tempstack);
		 if typecheck("bool",g) then
		   (if g = Bool(true) then 
		     (let old = newcl in 
		     let newl = 
		       (match labelcom cl with
			| Coml newl1 -> newl1
			| _ -> failwith("impossible in while")) in
		      let nuovo = Coml(newl @ [Com1(While(e, cl))] @ old) in
		      pop(continuation); push(nuovo,continuation))
		   else ())
		 else failwith ("nonboolean guard")		 
	   | Com1(Cifthenelse(e, cl1, cl2)) -> 
		pop(continuation); 
		push(Coml(Com2(Cifthenelse(e, cl1, cl2))::newcl),continuation);
		push(Expr1(e), continuation)		 
	   | Com2(Cifthenelse(e, cl1, cl2)) -> 
		 let g = top(tempstack) in
		 pop(tempstack);
		 if typecheck("bool",g) then
		    (let temp = if g = Bool(true) then
		    labelcom (cl1) else labelcom (cl2) in
		    let newl = 
		       (match temp with
			| Coml newl1 -> newl1
			| _ -> failwith("impossible in cifthenelse")) in
		    let nuovo = Coml(newl @ newcl) in
			   pop(continuation); push(nuovo,continuation))
		else failwith ("nonboolean guard")	
			   
	   | Com1(Call(e, el)) -> 
		pop(continuation); 
		push(Coml(Com2(Call(e, el))::newcl),continuation);
		push(Exprd1( e), continuation);
		pushargs(el, continuation)	   
	   | Com2(Call(e, el)) -> 
		 let p = top(tempdstack) in
		 pop(tempdstack);
		 let args = getargs(el,tempdstack) in		 
		 applyproc(p, args, sigma)  		     
	   | Com1(Block((l1, l2, l3))) -> 
		newframes(labelcom(l3), rho, sigma);    
		push(Rdecl(l2), top(cstack));
		push(labeldec(l1),top(cstack));
		pushlocalenv(l1,l2,rho);  (* nuovo *)
		pushlocalstore(l1);  (* nuovo *)
		()
	   | _ -> failwith("no more sensible cases in commands") ))

let initstate (r, s) = 
    svuota(cstack); svuota(tempvalstack); svuota(tempdvalstack);
    svuota(labelstack); currentenv := r; currentstore := s

	   
let loop () = 
   while not(empty(cstack)) do
    while not(empty(top(cstack))) do
     let currconstr = top(top(cstack)) in
	 (match currconstr with
	    | Expr1(e) -> itsem()
	    | Expr2(e) -> itsem()
	    | Exprd1(e) -> itsemden()
	    | Exprd2(e) -> itsemden()
	    | Coml(cl) -> itsemcl()
	    | Decl(l) -> itsemdecl()
	    | Rdecl(l) -> itsemrdecl()
	    | _ -> failwith("non legal construct in loop"))  
    done;
	(match top(labelstack) with
	  | Expr1(_) -> let valore = top(top(tempvalstack)) in
	       pop(top(tempvalstack));
		   pop(tempvalstack); push(valore,top(tempvalstack));
		   retainorpopenv(!initenv, valore); popstore(); pop(tempdvalstack)
	  | Exprd1(_) -> let valore = top(top(tempdvalstack)) in
	       pop(top(tempdvalstack));
		   pop(tempdvalstack); push(valore,top(tempdvalstack));
		   retainorpopenv(!initenv, dvaltoeval(valore)); popstore(); pop(tempvalstack)
	  | Decl(_) -> 
		   pop(tempvalstack); pop(tempdvalstack)
	  | Rdecl(_) -> 
		   pop(tempvalstack); pop(tempdvalstack)
	  | Coml(_) ->  pop(tempvalstack); pop(tempdvalstack);
	       popenv(); popstore()
	  | _ -> failwith("non legal label in loop"));
	pop(cstack);
    pop(labelstack)
  done

    
let sem (e,(r: dval env), (s: mval store)) = 
   initstate(r,s);
   initenv := r;  (* nuovo *)
   push(emptystack(tframesize(e),Novalue),tempvalstack);
   newframes(Expr1(e), r, s);
   loop();
   let valore= top(top(tempvalstack)) in
   pop(tempvalstack);  
   valore
   

let semden (e,(r: dval env), (s: mval store)) = 
   initstate(r,s);
   initenv := r;
   push(emptystack(tdframesize(e),Unbound),tempdvalstack);
   newframes(Exprd1(e), r, s);
   loop();
   let valore= top(top(tempdvalstack)) in
   pop(tempdvalstack); 
   valore
   
let semcl (cl,(r: dval env), (s: mval store)) = 
   initstate(r,s);
   newframes(labelcom(cl), r, s);
   loop();
   let st = topstore() in st 
   
let semc((c: com), (r:dval env), (s: mval store)) = 
   semcl ([c],r, s)

   


let semdv(dl, r, s) = 
   initstate(r,s);
   newframes(labeldec(dl), r, s);
   pushlocalenv(dl,[],!currentenv); (* nuovo *)
   pushlocalstore(dl);  (* nuovo *)
   loop();
   let st = topstore() in
   let rt = topenv() in
   (rt, st) 

let semdr(dl, r, s) = 
      initstate(r,s);
      newframes(Rdecl(dl), r, s);
      pushlocalenv([],dl,!currentenv);
      loop();
      let rt = topenv() in
      (rt, s) 
      
let semdl((dl,rl), r, s) = 
	 initstate(r,s);
	 newframes(Rdecl(rl), r, s);
	 push(labeldec(dl),top(cstack));
	 pushlocalenv(dl,rl,!currentenv);
	 pushlocalstore(dl);
	 loop();
	 let st = topstore() in
	 let rt = topenv() in
	 (rt, st)
	 

   
let semb ((l1, l2, l3), r, s) = initstate(r,s);
    newframes(labelcom(l3), r, s);    
		push(Rdecl(l2), top(cstack));
		push(labeldec(l1),top(cstack));
		pushlocalenv(l1,l2,!currentenv);  (* nuovo *)
		pushlocalstore(l1);  (* nuovo *)
		loop();
		currentenv := !currentenv + 1; (* nuovo *)
		currentstore := !currentstore + 1; (* nuovo *)
		topstore()





